# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""FloatingIP Mapping

Revision ID: fd98aa15958d

"""

# revision identifiers, used by Alembic.
revision = 'fd98aa15958d'
down_revision = '3791adbf0045'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gpm_pt_floatingip_mappings',
        sa.Column('policy_target_id', sa.String(length=36), nullable=False),
        sa.Column('floatingip_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('policy_target_id', 'floatingip_id'),
        sa.ForeignKeyConstraint(['policy_target_id'],
                                ['gp_policy_targets.id'],
                                ondelete='CASCADE',
                                name='gpm_pt_fip_map_fk_pt'),
        sa.ForeignKeyConstraint(['floatingip_id'],
                                ['floatingips.id'],
                                ondelete='CASCADE',
                                name='gpm_pt_fip_map_fk_fip'),
    )

    op.create_table(
        'gpm_service_policy_fip_mappings',
        sa.Column('service_policy_id', sa.String(length=36), nullable=False),
        sa.Column('policy_target_group_id',
                  sa.String(length=36),
                  nullable=False),
        sa.Column('floatingip_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('policy_target_group_id',
                                'service_policy_id',
                                'floatingip_id',
                                name='gpm_nsp_fip_map_pk_ptg_nsp_fip'),
        sa.ForeignKeyConstraint(['policy_target_group_id'],
                                ['gp_policy_target_groups.id'],
                                ondelete='CASCADE',
                                name='gpm_nsp_fip_map_fk_ptg'),
        sa.ForeignKeyConstraint(['service_policy_id'],
                                ['gp_network_service_policies.id'],
                                ondelete='CASCADE',
                                name='gpm_nsp_fip_map_fk_nsp'),
        sa.ForeignKeyConstraint(['floatingip_id'],
                                ['floatingips.id'],
                                ondelete='CASCADE',
                                name='gpm_nsp_fip_map_fk_fip'),
    )


def downgrade():

    op.drop_table('gpm_service_policy_fip_mappings')
    op.drop_table('gpm_pt_floatingip_mappings')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""port_erspan_extension

Revision ID: 016a678fafd4
Revises: bda3c34581e0
Create Date: 2020-11-03 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = '016a678fafd4'
down_revision = 'bda3c34581e0'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'apic_aim_port_erspan_configurations',
        sa.Column('port_id', sa.String(36), nullable=False),
        sa.Column('dest_ip', sa.String(64), nullable=False),
        sa.Column('flow_id', sa.Integer, nullable=False),
        sa.Column('direction', sa.Enum('in', 'out', 'both'), nullable=False),
        sa.ForeignKeyConstraint(
            ['port_id'], ['ports.id'],
            name='apic_aim_port_erspan_extensions_fk_port',
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id', 'dest_ip', 'flow_id', 'direction'))


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""network_migration_for_svi

Revision ID: 1c564e737f9f
Revises: 804d991a3564
Create Date: 2018-02-27 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = '1c564e737f9f'
down_revision = '804d991a3564'

from alembic import op
from alembic import util
from neutron.db import models_v2
import sqlalchemy as sa
from sqlalchemy.orm import lazyload


NetworkExtensionDb = sa.Table(
    'apic_aim_network_extensions', sa.MetaData(),
    sa.Column('network_id', sa.String(36), nullable=False),
    sa.Column('external_network_dn', sa.String(1024)),
    sa.Column('nat_type', sa.Enum('distributed', 'edge', '')),
    sa.Column('svi', sa.Boolean),
    sa.Column('bgp_enable', sa.Boolean,
              server_default=sa.false(), nullable=False),
    sa.Column('bgp_type', sa.Enum('default_export', ''),
              server_default="default_export", nullable=False),
    sa.Column('bgp_asn', sa.String(64),
              server_default="0", nullable=False))


def upgrade():
    session = sa.orm.Session(bind=op.get_bind(), autocommit=True)
    with session.begin(subtransactions=True):
        # Migrate networks.
        net_dbs = (session.query(models_v2.Network)
                   .options(lazyload('*')).all())
        for net_db in net_dbs:
            # Skip applying this update if using a legacy plugin
            if not hasattr(net_db, 'aim_extension_mapping') or (
                    net_db.aim_extension_mapping is None):
                continue
            util.msg("Migrating network: %s" % net_db)
            # If this update is successful then it means its an external
            # network with its DN set.
            res = session.execute(NetworkExtensionDb.update().values(
                svi=False).where(NetworkExtensionDb.c.network_id == net_db.id))
            if res.rowcount == 0:
                session.execute(NetworkExtensionDb.insert().values(
                    network_id=net_db.id, svi=False))


def downgrade():
    pass



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
import webob.exc

from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_grouppolicy_plugin as test_plugin)


class ImplicitPolicyTestCase(
        test_plugin.GroupPolicyPluginTestCase):

    def setUp(self):
        cfg.CONF.set_override('policy_drivers', ['implicit_policy'],
                              group='group_policy')
        super(ImplicitPolicyTestCase, self).setUp()


class TestImplicitL2Policy(ImplicitPolicyTestCase):

    def _test_implicit_lifecycle(self, shared=False):
        # Create policy_target group with implicit L2 policy.
        ptg1 = self.create_policy_target_group(shared=shared)
        self.assertEqual(shared, ptg1['policy_target_group']['shared'])
        ptg1_id = ptg1['policy_target_group']['id']
        l2p1_id = ptg1['policy_target_group']['l2_policy_id']
        self.assertIsNotNone(l2p1_id)
        req = self.new_show_request('policy_target_groups', ptg1_id,
                                    fmt=self.fmt)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual(l2p1_id, res['policy_target_group']['l2_policy_id'])
        req = self.new_show_request('l2_policies', l2p1_id, fmt=self.fmt)
        l2p1 = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual(ptg1['policy_target_group']['name'],
                         l2p1['l2_policy']['name'])
        self.assertEqual(shared, l2p1['l2_policy']['shared'])

        # Create 2nd policy_target group with different implicit L2 policy.
        ptg2 = self.create_policy_target_group()
        ptg2_id = ptg2['policy_target_group']['id']
        l2p2_id = ptg2['policy_target_group']['l2_policy_id']
        self.assertIsNotNone(l2p2_id)
        self.assertNotEqual(l2p1_id, l2p2_id)

        # Verify deleting 1st policy_target group does cleanup its L2
        # policy.
        req = self.new_delete_request('policy_target_groups', ptg1_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l2_policies', l2p1_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

        # Verify deleting 2nd policy_target group does cleanup its L2
        # policy.
        req = self.new_delete_request('policy_target_groups', ptg2_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l2_policies', l2p2_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_implicit_shared_lifecycle_negative(self):
        # Create PTG non shared
        self.create_policy_target_group()
        # This creates a non shared L3 policy that will be reused for any other
        # implicit PTG. Now create a shared PTG and verify that fails
        self.create_policy_target_group(
            shared=True, expected_res_status=webob.exc.HTTPBadRequest.code)

    def test_impicit_lifecycle(self):
        self._test_implicit_lifecycle()

    def test_implicit_lifecycle_shared(self):
        self._test_implicit_lifecycle(True)

    def test_explicit_lifecycle(self):
        # Create policy_target group with explicit L2 policy.
        l2p = self.create_l2_policy()
        l2p_id = l2p['l2_policy']['id']
        ptg = self.create_policy_target_group(l2_policy_id=l2p_id)
        ptg_id = ptg['policy_target_group']['id']
        self.assertEqual(l2p_id, ptg['policy_target_group']['l2_policy_id'])

        # Verify deleting policy_target group does not cleanup L2 policy.
        req = self.new_delete_request('policy_target_groups', ptg_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l2_policies', l2p_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_delete_from_implicit(self):
        # Create policy_target group with explicit L2 policy.
        l2p1 = self.create_l2_policy()
        l2p1_id = l2p1['l2_policy']['id']
        ptg = self.create_policy_target_group(l2_policy_id=l2p1_id)
        ptg_id = ptg['policy_target_group']['id']
        self.assertEqual(l2p1_id, ptg['policy_target_group']['l2_policy_id'])

        # Delete PTG
        self.delete_policy_target_group(ptg_id, expected_res_status=204)

        # Verify old L2 policy was not cleaned up.
        req = self.new_show_request('l2_policies', l2p1_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)


class TestImplicitL3Policy(ImplicitPolicyTestCase):

    def _test_implicit_lifecycle(self, shared=False):
        # Create L2 policy with implicit L3 policy.
        l2p1 = self.create_l2_policy(shared=shared)
        l2p1_id = l2p1['l2_policy']['id']
        l3p_id = l2p1['l2_policy']['l3_policy_id']
        self.assertIsNotNone(l3p_id)
        self.assertEqual(shared, l2p1['l2_policy']['shared'])
        req = self.new_show_request('l2_policies', l2p1_id, fmt=self.fmt)
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual(l3p_id, res['l2_policy']['l3_policy_id'])
        req = self.new_show_request('l3_policies', l3p_id, fmt=self.fmt)
        l3p = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual('default', l3p['l3_policy']['name'])
        self.assertEqual(shared, l3p['l3_policy']['shared'])
        self.assertEqual(24, l3p['l3_policy']['subnet_prefix_length'])

        # Create 2nd L2 policy sharing implicit L3 policy.
        l2p2 = self.create_l2_policy()
        l2p2_id = l2p2['l2_policy']['id']
        self.assertEqual(l3p_id, l2p2['l2_policy']['l3_policy_id'])

        # Verify deleting 1st L2 policy does not cleanup L3 policy.
        req = self.new_delete_request('l2_policies', l2p1_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

        # Verify deleting last L2 policy does cleanup L3 policy.
        req = self.new_delete_request('l2_policies', l2p2_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_impicit_lifecycle(self):
        self._test_implicit_lifecycle()

    def test_implicit_lifecycle_shared(self):
        self._test_implicit_lifecycle(True)

    def test_explicit_lifecycle(self):
        # Create L2 policy with explicit L3 policy.
        l3p = self.create_l3_policy()
        l3p_id = l3p['l3_policy']['id']
        l2p = self.create_l2_policy(l3_policy_id=l3p_id)
        l2p_id = l2p['l2_policy']['id']
        self.assertEqual(l3p_id, l2p['l2_policy']['l3_policy_id'])

        # Verify deleting L2 policy does not cleanup L3 policy.
        req = self.new_delete_request('l2_policies', l2p_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_unowned_default_lifecycle(self):
        # Create L2 policy with unowned default L3 policy.
        l3p = self.create_l3_policy(name='default')
        l3p_id = l3p['l3_policy']['id']
        l2p = self.create_l2_policy()
        l2p_id = l2p['l2_policy']['id']
        self.assertEqual(l3p_id, l2p['l2_policy']['l3_policy_id'])

        # Verify deleting L2 policy does not cleanup L3 policy.
        req = self.new_delete_request('l2_policies', l2p_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_single_default_policy(self):
        # Verify only one default L3 policy can be created per tenant.
        l3p = self.create_l3_policy(name='default')
        self.assertEqual('default', l3p['l3_policy']['name'])
        res = self.create_l3_policy(name='default', expected_res_status=400)
        self.assertEqual('DefaultL3PolicyAlreadyExists',
                         res['NeutronError']['type'])

    def test_update_from_implicit(self):
        # Create L2 policy with implicit L3 policy.
        l2p = self.create_l2_policy()
        l2p_id = l2p['l2_policy']['id']
        l3p1_id = l2p['l2_policy']['l3_policy_id']
        req = self.new_show_request('l3_policies', l3p1_id, fmt=self.fmt)
        l3p1 = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual('default', l3p1['l3_policy']['name'])

        # Update L2 policy to explicit L3 policy.
        l3p2 = self.create_l3_policy()
        l3p2_id = l3p2['l3_policy']['id']
        data = {'l2_policy': {'l3_policy_id': l3p2_id}}
        req = self.new_update_request('l2_policies', data, l2p_id)
        l2p = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual(l3p2_id, l2p['l2_policy']['l3_policy_id'])

        # Verify old L3 policy was cleaned up
        req = self.new_show_request('l3_policies', l3p1_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

        # Verify deleting L2 policy does not cleanup new L3 policy.
        req = self.new_delete_request('l2_policies', l2p_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p2_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

    def test_update_to_implicit(self):
        # Create L2 policy with explicit L3 policy.
        l3p1 = self.create_l3_policy()
        l3p1_id = l3p1['l3_policy']['id']
        l2p = self.create_l2_policy(l3_policy_id=l3p1_id)
        l2p_id = l2p['l2_policy']['id']
        self.assertEqual(l3p1_id, l2p['l2_policy']['l3_policy_id'])

        # Update L2 policy to implicit L3 policy.
        data = {'l2_policy': {'l3_policy_id': None}}
        req = self.new_update_request('l2_policies', data, l2p_id)
        l2p = self.deserialize(self.fmt, req.get_response(self.ext_api))
        l3p2_id = l2p['l2_policy']['l3_policy_id']
        self.assertNotEqual(l3p1_id, l3p2_id)
        self.assertIsNotNone(l3p2_id)
        req = self.new_show_request('l3_policies', l3p2_id, fmt=self.fmt)
        l3p2 = self.deserialize(self.fmt, req.get_response(self.ext_api))
        self.assertEqual('default', l3p2['l3_policy']['name'])

        # Verify old L3 policy was not cleaned up.
        req = self.new_show_request('l3_policies', l3p1_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPOk.code, res.status_int)

        # Verify deleting L2 policy does cleanup new L3 policy.
        req = self.new_delete_request('l2_policies', l2p_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)
        req = self.new_show_request('l3_policies', l3p2_id, fmt=self.fmt)
        res = req.get_response(self.ext_api)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)


class TestImplicitExternalSegment(ImplicitPolicyTestCase):

    def setUp(self):
        self._default_es_name = 'default'
        super(TestImplicitExternalSegment, self).setUp()
        cfg.CONF.set_override(
            'default_external_segment_name', self._default_es_name,
            group='group_policy_implicit_policy')

    def _create_default_es(self, **kwargs):
        return self.create_external_segment(name=self._default_es_name,
                                            **kwargs)

    def _test_implicit_lifecycle(self, shared=False):
        # Create default ES
        es = self._create_default_es(shared=shared)['external_segment']
        # Create non-default ES
        ndes = self.create_external_segment(
            name='non-default-name')['external_segment']

        # Create EP without ES set
        ep = self.create_external_policy()['external_policy']
        self.assertEqual(es['id'], ep['external_segments'][0])
        # Verify persisted
        req = self.new_show_request('external_policies', ep['id'],
                                    fmt=self.fmt)
        ep = self.deserialize(
            self.fmt, req.get_response(self.ext_api))['external_policy']
        self.assertEqual(es['id'], ep['external_segments'][0])

        # Verify update
        ep = self.update_external_policy(
            ep['id'], expected_res_status=200,
            external_segments=[ndes['id']])['external_policy']
        self.assertEqual(ndes['id'], ep['external_segments'][0])
        self.assertEqual(1, len(ep['external_segments']))

        # Create L3P without ES set
        l3p = self.create_l3_policy()['l3_policy']
        self.assertEqual(es['id'], list(l3p['external_segments'].keys())[0])
        # Verify persisted
        req = self.new_show_request('l3_policies', l3p['id'],
                                    fmt=self.fmt)
        l3p = self.deserialize(
            self.fmt, req.get_response(self.ext_api))['l3_policy']
        self.assertEqual(es['id'], list(l3p['external_segments'].keys())[0])

        # Verify update
        l3p = self.update_l3_policy(
            l3p['id'], expected_res_status=200,
            external_segments={ndes['id']: []})['l3_policy']
        self.assertEqual(ndes['id'], list(l3p['external_segments'].keys())[0])
        self.assertEqual(1, len(l3p['external_segments']))

        # Verify only one visible ES can exist
        res = self._create_default_es(expected_res_status=400)
        self.assertEqual('DefaultExternalSegmentAlreadyExists',
                         res['NeutronError']['type'])

    def test_impicit_lifecycle(self):
        self._test_implicit_lifecycle()

    def test_implicit_lifecycle_shared(self):
        self._test_implicit_lifecycle(True)

    def test_implicit_shared_visibility(self):
        es = self._create_default_es(shared=True,
                                     tenant_id='onetenant')['external_segment']
        ep = self.create_external_policy(
            tenant_id='anothertenant')['external_policy']
        self.assertEqual(es['id'], ep['external_segments'][0])
        self.assertEqual(1, len(ep['external_segments']))

        l3p = self.create_l3_policy(
            tenant_id='anothertenant')['l3_policy']
        self.assertEqual(es['id'], list(l3p['external_segments'].keys())[0])
        self.assertEqual(1, len(ep['external_segments']))

        res = self._create_default_es(expected_res_status=400,
                                      tenant_id='anothertenant')
        self.assertEqual('DefaultExternalSegmentAlreadyExists',
                         res['NeutronError']['type'])


class TestQuotasForGBPWithImplicitDriver(ImplicitPolicyTestCase):

    def setUp(self):
        cfg.CONF.set_override('quota_l3_policy', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_l2_policy', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target_group', 1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target', 1, group='QUOTAS')
        super(TestQuotasForGBPWithImplicitDriver, self).setUp()

    def tearDown(self):
        cfg.CONF.set_override('quota_l3_policy', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_l2_policy', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target_group', -1, group='QUOTAS')
        cfg.CONF.set_override('quota_policy_target', -1, group='QUOTAS')
        super(TestQuotasForGBPWithImplicitDriver, self).tearDown()

    def test_quota_for_group_resources_implicit(self):
        # The following tests that implicitly created L2P and L3P
        # are counted as a part of the resource quota.
        ptg_id = self.create_policy_target_group()['policy_target_group']['id']
        self.create_policy_target(policy_target_group_id=ptg_id)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target,
                          policy_target_group_id=ptg_id)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target_group)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_l3_policy)
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_l2_policy)

    def test_quota_for_implicit_l3p(self):
        # The following tests that implicit L3P creation fails
        # when resource quota is reached.
        cfg.CONF.set_override('quota_policy_target_group', 2, group='QUOTAS')
        cfg.CONF.set_override('quota_l2_policy', 2, group='QUOTAS')
        l3p = self.create_l3_policy(name='test')
        l3p_id = l3p['l3_policy']['id']
        l2p = self.create_l2_policy(name='test', l3_policy_id=l3p_id)
        l2p_id = l2p['l2_policy']['id']
        self.create_policy_target_group(
            l2_policy_id=l2p_id)['policy_target_group']['id']
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target_group)

    def test_quota_for_implicit_l2p(self):
        # The following tests that implicit L2P creation fails
        # when resource quota is reached.
        cfg.CONF.set_override('quota_policy_target_group', 2, group='QUOTAS')
        cfg.CONF.set_override('quota_l3_policy', 2, group='QUOTAS')
        l3p = self.create_l3_policy(name='test')
        l3p_id = l3p['l3_policy']['id']
        l2p = self.create_l2_policy(name='test', l3_policy_id=l3p_id)
        l2p_id = l2p['l2_policy']['id']
        self.create_policy_target_group(
            l2_policy_id=l2p_id)['policy_target_group']['id']
        self.assertRaises(webob.exc.HTTPClientError,
                          self.create_policy_target_group)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Active active AAP

Revision ID: b76dc22f2e23
Revises: 24427b3a5c95
Create Date: 2019-11-21 14:18:11.909757

"""

# revision identifiers, used by Alembic.
revision = 'b76dc22f2e23'
down_revision = '24427b3a5c95'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('apic_aim_subnet_extensions',
                  sa.Column('active_active_aap', sa.Boolean,
                            server_default=sa.false(), nullable=False))


def downgrade():
    pass



# Copyright (c) 2018 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from neutron.api.v2 import router
from neutron.common import config
from neutron import manager
from neutron_lib.plugins import directory

from gbpservice._i18n import _
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

# Any policy-driver specific CLI options must be included here, since
# the CLI options must be registered before the GBP service plugin and
# the configured policy drivers can be loaded.
cli_opts = [
    cfg.BoolOpt('repair', default=False, help='Enable repair of invalid state.'),
    cfg.ListOpt('resources', default=[], help='List of resources to be reconciled. '
    'Avaiable options are router, security_group, network, port, subnetpool, floatingip, address_scope'),
    cfg.ListOpt('tenants', default=[], help='List of tenants to be reconciled')
]


def main():
    cfg.CONF.register_cli_opts(cli_opts)
    config.init(sys.argv[1:])

    # Enable logging but prevent output to stderr.
    cfg.CONF.use_stderr = False
    config.setup_logging()

    if not cfg.CONF.config_file:
        sys.exit(_("ERROR: Unable to find configuration file via the default"
                   " search paths (~/.neutron/, ~/, /etc/neutron/, /etc/) and"
                   " the '--config-file' option!"))

    router.APIRouter.factory({})
    manager.init()

    gbp_plugin = directory.get_plugin('GROUP_POLICY')
    if not gbp_plugin:
        sys.exit("GBP service plugin not configured.")

    result = gbp_plugin.validate_state(cfg.CONF.repair, cfg.CONF.resources, cfg.CONF.tenants)
    if result in [api.VALIDATION_FAILED_REPAIRABLE,
                  api.VALIDATION_FAILED_UNREPAIRABLE,
                  api.VALIDATION_FAILED_WITH_EXCEPTION]:
        sys.exit(result)
    return 0



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
import six

from neutron.common import utils as n_utils
from neutron.quota import resource_registry
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib import context as n_ctx
from neutron_lib.plugins import constants as pconst
from neutron_lib.plugins import directory
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_utils import excutils

from gbpservice.common import utils as gbp_utils
from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.db.grouppolicy import group_policy_db as gpdb
from gbpservice.neutron.db.grouppolicy import group_policy_mapping_db
from gbpservice.neutron import extensions as gbp_extensions
from gbpservice.neutron.extensions import group_policy as gpex
from gbpservice.neutron.services.grouppolicy import (
    extension_manager as ext_manager)
from gbpservice.neutron.services.grouppolicy import (
    group_policy_context as p_context)
from gbpservice.neutron.services.grouppolicy import (
    policy_driver_manager as manager)
from gbpservice.neutron.services.grouppolicy.common import constants as gp_cts
from gbpservice.neutron.services.grouppolicy.common import exceptions as gp_exc
from gbpservice.neutron.services.grouppolicy.common import utils
from gbpservice.neutron.services.servicechain.plugins.ncp import (
    model as ncp_model)


LOG = logging.getLogger(__name__)
STATUS = 'status'
STATUS_DETAILS = 'status_details'
STATUS_SET = set([STATUS, STATUS_DETAILS])


class GroupPolicyPlugin(group_policy_mapping_db.GroupPolicyMappingDbPlugin):

    """Implementation of the Group Policy Model Plugin.

    This class manages the workflow of Group Policy request/response.
    Most DB related works are implemented in class
    db_group_policy_mapping.GroupPolicyMappingDbMixin.
    """
    _supported_extension_aliases = ["group-policy", "group-policy-mapping"]
    path_prefix = gp_cts.GBP_PREFIXES[pconst.GROUP_POLICY]

    @property
    def supported_extension_aliases(self):
        if not hasattr(self, '_aliases'):
            aliases = self._supported_extension_aliases[:]
            aliases += self.extension_manager.extension_aliases()
            self._aliases = aliases
        return self._aliases

    def start_rpc_listeners(self):
        return self.policy_driver_manager.start_rpc_listeners()

    def validate_state(self, repair, resources, tenants):
        return self.policy_driver_manager.validate_state(repair,
            resources, tenants)

    @property
    def servicechain_plugin(self):
        # REVISIT(rkukura): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        servicechain_plugin = directory.get_plugin(pconst.SERVICECHAIN)
        if not servicechain_plugin:
            LOG.error("No Servicechain service plugin found.")
            raise gp_exc.GroupPolicyDeploymentError()
        return servicechain_plugin

    # Shared attribute validation rules:
    # - A shared resource cannot use/link a non-shared resource
    # - A shared resource cannot be reverted to non-shared if used/linked by
    # other shared resources, or by any resource owned by any other tenant

    # In the usage graph, specify which resource has to be checked to validate
    # sharing policy conformity:
    # usage_graph = {<to_check>: {<attribute>: <type>}, ...}
    # <attribute> is the field on the <to_check> dictionary that can be used
    # to retrieve the UUID/s of the specific object <type>

    usage_graph = {'l3_policy': {'external_segments':
                                 'external_segment'},
                   'l2_policy': {'l3_policy_id': 'l3_policy'},
                   'policy_target_group': {
                       'network_service_policy_id': 'network_service_policy',
                       'l2_policy_id': 'l2_policy',
                       'provided_policy_rule_sets': 'policy_rule_set',
                       'consumed_policy_rule_sets': 'policy_rule_set'},
                   'network_service_policy': {},
                   'policy_rule': {
                       'policy_classifier_id': 'policy_classifier',
                       'policy_actions': 'policy_action'},
                   'policy_action': {},
                   'policy_classifier': {},
                   'policy_rule_set': {
                       'parent_id': 'policy_rule_set',
                       'policy_rules': 'policy_rule'},
                   'external_segment': {},
                   'external_policy': {
                       'external_segments': 'external_segment',
                       'provided_policy_rule_sets': 'policy_rule_set',
                       'consumed_policy_rule_sets': 'policy_rule_set'},
                   'nat_pool': {'external_segment_id':
                                'external_segment'},
                   'policy_target': {'policy_target_group_id':
                                     'policy_target_group'},
                   'application_policy_group': {}
                   }

    @staticmethod
    def _validate_shared_create(self, context, obj, identity):
        # REVISIT(ivar): only validate new references
        links = self.usage_graph.get(identity, {})
        for attr in links:
            ids = obj[attr]
            if ids:
                if isinstance(ids, six.string_types):
                    ids = [ids]
                ref_type = links[attr]
                linked_objects = getattr(
                    self, 'get_%s' % gbp_extensions.get_plural(ref_type))(
                        context, filters={'id': ids})
                link_ids = set()
                for linked in linked_objects:
                    link_ids.add(linked['id'])
                    GroupPolicyPlugin._verify_sharing_consistency(
                        obj, linked, identity, ref_type, context.is_admin)
                # Check for missing references
                missing = set(ids) - link_ids
                if missing:
                    raise gpex.GbpResourceNotFound(identity=ref_type,
                                                   id=str(missing))

    @staticmethod
    def _validate_shared_update(self, context, original, updated, identity):
        # Need admin context to check sharing constraints

        # Even though the shared attribute may not be changed, the objects
        # it is referring to might. For this reson we run the reference
        # validation every time a shared resource is updated
        # TODO(ivar): run only when relevant updates happen
        self._validate_shared_create(self, context, updated, identity)
        if updated.get('shared') != original.get('shared'):
            context = context.elevated()
            getattr(self, '_validate_%s_unshare' % identity)(context, updated)

    @staticmethod
    def _check_shared_or_different_tenant(context, obj, method, attr,
                                          value=None):
        tenant_id = obj['tenant_id']
        refs = method(context, filters={attr: value or [obj['id']]})
        for ref in refs:
            if ref.get('shared') or tenant_id != ref['tenant_id']:
                raise gp_exc.InvalidSharedAttributeUpdate(id=obj['id'],
                                                          rid=ref['id'])

    def _validate_l3_policy_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_l2_policies, 'l3_policy_id')

    def _validate_l2_policy_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_target_groups, 'l2_policy_id')

    def _validate_policy_target_group_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_targets, 'policy_target_group_id')

    def _validate_network_service_policy_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_target_groups,
            'network_service_policy_id')

    def _validate_policy_rule_set_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_target_groups, 'id',
            obj['providing_policy_target_groups'] +
            obj['consuming_policy_target_groups'])
        self._check_shared_or_different_tenant(
            context, obj, self.get_external_policies, 'id',
            obj['providing_external_policies'] +
            obj['consuming_external_policies'])

    def _validate_policy_classifier_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_rules, 'policy_classifier_id')

    def _validate_policy_rule_unshare(self, context, obj):
        c_ids = self._get_policy_rule_policy_rule_sets(context, obj['id'])
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_rule_sets, 'id', c_ids)

    def _validate_policy_action_unshare(self, context, obj):
        r_ids = self._get_policy_action_rules(context, obj['id'])
        self._check_shared_or_different_tenant(
            context, obj, self.get_policy_rules, 'id', r_ids)

    def _validate_external_segment_unshare(self, context, obj):
        self._check_shared_or_different_tenant(
            context, obj, self.get_l3_policies, 'id', obj['l3_policies'])
        self._check_shared_or_different_tenant(
            context, obj, self.get_external_policies, 'id',
            obj['external_policies'])
        self._check_shared_or_different_tenant(
            context, obj, self.get_nat_pools, 'external_segment_id')

    def _validate_external_policy_unshare(self, context, obj):
        pass

    def _validate_nat_pool_unshare(self, context, obj):
        pass

    def _validate_routes(self, context, current, original=None):
        if original:
            added = (set((x['destination'], x['nexthop']) for x in
                         current['external_routes']) -
                     set((x['destination'], x['nexthop']) for x in
                         original['external_routes']))
        else:
            added = set((x['destination'], x['nexthop']) for x in
                        current['external_routes'])
        if added:
            # Verify new ones don't overlap with the existing L3P
            added_dest = set(x[0] for x in added)
            # Remove default routes
            added_dest.discard('0.0.0.0/0')
            added_dest.discard('::/0')
            added_ipset = netaddr.IPSet(added_dest)
            if current['l3_policies']:
                l3ps = self.get_l3_policies(
                    context, filters={'id': current['l3_policies']})
                for l3p in l3ps:
                    ip_pool_list = utils.convert_ip_pool_string_to_list(
                        l3p['ip_pool'])
                    if netaddr.IPSet(ip_pool_list) & added_ipset:
                        raise gp_exc.ExternalRouteOverlapsWithL3PIpPool(
                            destination=added_dest, l3p_id=l3p['id'],
                            es_id=current['id'])
                    es_list = [current]
                    es_list.extend(self.get_external_segments(
                        context.elevated(),
                        filters={'id': [e for e in l3p['external_segments']
                                        if e != current['id']]}))
                    self._validate_identical_external_routes(es_list)
            # Verify NH in ES pool
            added_nexthop = netaddr.IPSet(x[1] for x in added if x[1])
            es_subnet = netaddr.IPSet([current['cidr']])
            if added_nexthop & es_subnet != added_nexthop:
                raise gp_exc.ExternalRouteNextHopNotInExternalSegment(
                    cidr=current['cidr'])

    def _validate_l3p_es(self, context, current, original=None):
        if original:
            added = (set(current['external_segments'].keys()) -
                     set(original['external_segments'].keys()))
        else:
            added = set(current['external_segments'].keys())
        if added:
            es_list = self.get_external_segments(context,
                                                 filters={'id': added})
            ip_pool_list = utils.convert_ip_pool_string_to_list(
                current['ip_pool'])
            l3p_ipset = netaddr.IPSet(ip_pool_list)
            for es in es_list:
                # Verify no route overlap
                dest_set = set(x['destination'] for x in
                               es['external_routes'])
                dest_set.discard('0.0.0.0/0')
                dest_set.discard('::/0')
                if l3p_ipset & netaddr.IPSet(dest_set):
                    raise gp_exc.ExternalRouteOverlapsWithL3PIpPool(
                        destination=dest_set, l3p_id=current['id'],
                        es_id=es['id'])
                # Verify segment CIDR doesn't overlap with L3P's
                cidr = es['cidr']
                if es['subnet_id']:
                    core_plugin = directory.get_plugin()
                    cidr = core_plugin.get_subnet(context,
                        es['subnet_id'])['cidr']
                if l3p_ipset & netaddr.IPSet([cidr]):
                    raise gp_exc.ExternalSegmentSubnetOverlapsWithL3PIpPool(
                        subnet=cidr, l3p_id=current['id'],
                        es_id=current['id'])
                # Verify allocated address correctly in subnet
                for addr in current['external_segments'][es['id']]:
                    if addr != gpdb.ADDRESS_NOT_SPECIFIED:
                        if addr not in netaddr.IPNetwork(cidr):
                            raise gp_exc.InvalidL3PExternalIPAddress(
                                ip=addr, es_id=es['id'], l3p_id=current['id'],
                                es_cidr=cidr)
            es_list_all = self.get_external_segments(
                context.elevated(),
                filters={'id': list(current['external_segments'].keys())})
            self._validate_identical_external_routes(es_list_all)

    def _validate_identical_external_routes(self, es_list):
        if len(es_list) < 2:
            return
        route_dict = {netaddr.IPNetwork(route['destination']).cidr: es
                      for es in es_list[1:]
                      for route in es['external_routes']}
        for route in es_list[0]['external_routes']:
            cidr = netaddr.IPNetwork(route['destination']).cidr
            if cidr in route_dict:
                raise gp_exc.IdenticalExternalRoute(
                    es1=es_list[0]['id'], es2=route_dict[cidr]['id'],
                    cidr=cidr)

    def _validate_action_value(self, context, action):
        if action.get('action_type') == gp_cts.GP_ACTION_REDIRECT:
            if action.get('action_value'):
                # Verify sc spec existence and visibility
                spec = self.servicechain_plugin.get_servicechain_spec(
                    context, action['action_value'])
                GroupPolicyPlugin._verify_sharing_consistency(
                    action, spec, 'policy_action', 'servicechain_spec',
                    context.is_admin)

    @staticmethod
    def _verify_sharing_consistency(primary, reference, primary_type,
                                    reference_type, is_admin):
        if not reference.get('shared'):
            if primary.get('shared'):
                raise gp_exc.SharedResourceReferenceError(
                    res_type=primary_type, res_id=primary['id'],
                    ref_type=reference_type, ref_id=reference['id'])
            if not is_admin:
                if primary.get('tenant_id') != reference.get('tenant_id'):
                    raise gp_exc.InvalidCrossTenantReference(
                        res_type=primary_type, res_id=primary['id'],
                        ref_type=reference_type, ref_id=reference['id'])

    def _get_status_from_drivers(self, context, context_name, resource_name,
                                 resource_id, resource):
        status = resource['status']
        status_details = resource['status_details']
        policy_context = getattr(p_context, context_name)(
            self, context, resource, resource)
        getattr(self.policy_driver_manager,
                "get_" + resource_name + "_status")(policy_context)
        _resource = getattr(policy_context, "_" + resource_name)
        updated_status = _resource['status']
        updated_status_details = _resource['status_details']
        if status != updated_status or (
                    status_details != updated_status_details):
            new_status = {resource_name: {'status': updated_status,
                                          'status_details':
                                          updated_status_details}}
            with db_api.CONTEXT_WRITER.using(context):
                getattr(super(GroupPolicyPlugin, self),
                        "update_" + resource_name)(
                            context, _resource['id'], new_status)
            resource['status'] = updated_status
            resource['status_details'] = updated_status_details
        return resource

    def _get_resource(self, context, resource_name, resource_id,
                      gbp_context_name, fields=None):
        # The following is a writer because we do DB write for status
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            get_method = "".join(['get_', resource_name])
            result = getattr(super(GroupPolicyPlugin, self), get_method)(
                context, resource_id, None)
            extend_resources_method = "".join(['extend_', resource_name,
                                               '_dict'])
            getattr(self.extension_manager, extend_resources_method)(
                session, result)

            # Invoke drivers only if status attributes are requested
            if not fields or STATUS_SET.intersection(set(fields)):
                result = self._get_status_from_drivers(
                    context, gbp_context_name, resource_name, resource_id,
                    result)
            return db_api.resource_fields(result, fields)

    def _get_resources(self, context, resource_name, gbp_context_name,
                       filters=None, fields=None, sorts=None, limit=None,
                       marker=None, page_reverse=False):
        # The following is a writer because we do DB write for status
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            resource_plural = gbp_utils.get_resource_plural(resource_name)
            get_resources_method = "".join(['get_', resource_plural])
            results = getattr(super(GroupPolicyPlugin, self),
                              get_resources_method)(
                context, filters, None, sorts, limit, marker, page_reverse)
            filtered_results = []
            for result in results:
                extend_resources_method = "".join(['extend_', resource_name,
                                                   '_dict'])
                getattr(self.extension_manager, extend_resources_method)(
                    session, result)
                filtered = self._filter_extended_result(result, filters)
                if filtered:
                    filtered_results.append(filtered)

        new_filtered_results = []
        # Invoke drivers only if status attributes are requested
        if not fields or STATUS_SET.intersection(set(fields)):
            for result in filtered_results:
                result = self._get_status_from_drivers(
                    context, gbp_context_name, resource_name, result['id'],
                    result)
                new_filtered_results.append(result)
        new_filtered_results = new_filtered_results or filtered_results
        return [db_api.resource_fields(nfresult, fields) for nfresult in
                new_filtered_results]

    @resource_registry.tracked_resources(
        l3_policy=group_policy_mapping_db.L3PolicyMapping,
        l2_policy=group_policy_mapping_db.L2PolicyMapping,
        policy_target=group_policy_mapping_db.PolicyTargetMapping,
        policy_target_group=group_policy_mapping_db.PolicyTargetGroupMapping,
        application_policy_group=gpdb.ApplicationPolicyGroup,
        policy_classifier=gpdb.PolicyClassifier,
        policy_action=gpdb.PolicyAction,
        policy_rule=gpdb.PolicyRule,
        policy_rule_set=gpdb.PolicyRuleSet,
        external_policy=gpdb.ExternalPolicy,
        external_segment=group_policy_mapping_db.ExternalSegmentMapping,
        nat_pool=group_policy_mapping_db.NATPoolMapping,
        network_service_policy=gpdb.NetworkServicePolicy)
    def __init__(self):
        self.extension_manager = ext_manager.ExtensionManager()
        self.policy_driver_manager = manager.PolicyDriverManager()
        super(GroupPolicyPlugin, self).__init__()
        self.extension_manager.initialize()
        self.policy_driver_manager.initialize()

    def _filter_extended_result(self, result, filters):
        filters = filters or {}
        for field in filters:
            # Ignore unknown fields
            if field in result:
                if result[field] not in filters[field]:
                    break
        else:
            return result

    def _add_fixed_ips_to_port_attributes(self, policy_target):
        if 'fixed_ips' in policy_target['policy_target'] and (
                policy_target['policy_target']['fixed_ips'] is not (
                    constants.ATTR_NOT_SPECIFIED)):
            port_attributes = {'fixed_ips': policy_target[
                'policy_target']['fixed_ips']}
            policy_target['policy_target'].update(
                {'port_attributes': port_attributes})

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_target(self, context, policy_target):
        self._ensure_tenant(context, policy_target['policy_target'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            self._add_fixed_ips_to_port_attributes(policy_target)
            result = super(GroupPolicyPlugin,
                           self).create_policy_target(context, policy_target)
            self.extension_manager.process_create_policy_target(
                session, policy_target, result)
            self._validate_shared_create(
                self, context, result, 'policy_target')
            policy_context = p_context.PolicyTargetContext(self, context,
                                                           result)
            self.policy_driver_manager.create_policy_target_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_target_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_policy_target_postcommit "
                              "failed, deleting policy_target %s",
                              result['id'])
                self.delete_policy_target(context, result['id'])

        return self.get_policy_target(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_target(self, context, policy_target_id, policy_target):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            self._add_fixed_ips_to_port_attributes(policy_target)
            original_policy_target = self.get_policy_target(context,
                                                            policy_target_id)
            updated_policy_target = super(
                GroupPolicyPlugin, self).update_policy_target(
                    context, policy_target_id, policy_target)
            self.extension_manager.process_update_policy_target(
                session, policy_target, updated_policy_target)
            self._validate_shared_update(self, context, original_policy_target,
                                         updated_policy_target,
                                         'policy_target')
            policy_context = p_context.PolicyTargetContext(
                self, context, updated_policy_target,
                original_policy_target=original_policy_target)
            self.policy_driver_manager.update_policy_target_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_target_postcommit(
            policy_context)
        return self.get_policy_target(context, policy_target_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_target(self, context, policy_target_id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_target = self.get_policy_target(context, policy_target_id)
            policy_context = p_context.PolicyTargetContext(
                self, context, policy_target)
            self.policy_driver_manager.delete_policy_target_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_policy_target(
                context, policy_target_id)

        try:
            self.policy_driver_manager.delete_policy_target_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_target_postcommit failed "
                          "for policy_target %s",
                          policy_target_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_target(self, context, policy_target_id, fields=None):
        return self._get_resource(context, 'policy_target', policy_target_id,
                                  'PolicyTargetContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_targets(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        return self._get_resources(
            context, 'policy_target', 'PolicyTargetContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_target_group(self, context, policy_target_group):
        self._ensure_tenant(context,
                            policy_target_group['policy_target_group'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_policy_target_group(
                               context, policy_target_group)
            self.extension_manager.process_create_policy_target_group(
                session, policy_target_group, result)
            self._validate_shared_create(self, context, result,
                                         'policy_target_group')
            policy_context = p_context.PolicyTargetGroupContext(
                self, context, result)
            self.policy_driver_manager.create_policy_target_group_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_target_group_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_policy_target_group_postcommit "
                              "failed, deleting policy_target_group %s",
                              result['id'])
                self.delete_policy_target_group(context, result['id'])

        return self.get_policy_target_group(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_target_group(self, context, policy_target_group_id,
                                   policy_target_group):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_policy_target_group = self.get_policy_target_group(
                    context, policy_target_group_id)
            updated_policy_target_group = super(
                GroupPolicyPlugin, self).update_policy_target_group(
                    context, policy_target_group_id, policy_target_group)
            # REVISIT(rkukura): We could potentially allow updates to
            # l2_policy_id when no policy targets exist. This would
            # involve removing each old subnet from the l3_policy's
            # router, deleting each old subnet, creating a new subnet on
            # the new l2_policy's network, and adding that subnet to the
            # l3_policy's router in postcommit. Its also possible that new
            # subnet[s] would be provided explicitly as part of the
            # update.
            old_l2p = original_policy_target_group['l2_policy_id']
            new_l2p = updated_policy_target_group['l2_policy_id']
            if old_l2p and old_l2p != new_l2p:
                raise gp_exc.L2PolicyUpdateOfPolicyTargetGroupNotSupported()

            self.extension_manager.process_update_policy_target_group(
                session, policy_target_group, updated_policy_target_group)
            self._validate_shared_update(
                self, context, original_policy_target_group,
                updated_policy_target_group, 'policy_target_group')
            policy_context = p_context.PolicyTargetGroupContext(
                self, context, updated_policy_target_group,
                original_policy_target_group=original_policy_target_group)
            self.policy_driver_manager.update_policy_target_group_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_target_group_postcommit(
            policy_context)

        return self.get_policy_target_group(context, policy_target_group_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_target_group(self, context, policy_target_group_id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_target_group = self.get_policy_target_group(
                context, policy_target_group_id)
            pt_ids = policy_target_group['policy_targets']
            for pt in self.get_policy_targets(context.elevated(),
                                              {'id': pt_ids}):
                if (pt['port_id'] and self._is_port_bound(pt['port_id']) and
                    not (self._is_service_target(context, pt['id']))):
                    raise gp_exc.PolicyTargetGroupInUse(
                        policy_target_group=policy_target_group_id)
            policy_context = p_context.PolicyTargetGroupContext(
                self, context, policy_target_group)
            self.policy_driver_manager.delete_policy_target_group_precommit(
                policy_context)

        # Disassociate all the PRSs first, this will trigger service chains
        # deletion.
        self.update_policy_target_group(
            context, policy_target_group_id,
            {'policy_target_group': {'provided_policy_rule_sets': {},
                                     'consumed_policy_rule_sets': {}}})
        policy_context.current['provided_policy_rule_sets'] = []
        policy_context.current['consumed_policy_rule_sets'] = []

        # Proxy PTGs must be deleted before the group itself.
        if policy_target_group.get('proxy_group_id'):
            try:
                self.delete_policy_target_group(
                    context, policy_target_group['proxy_group_id'])
            except gpex.PolicyTargetGroupNotFound:
                LOG.warning('PTG %s already deleted',
                            policy_target_group['proxy_group_id'])

        # Delete the PTG's PTs, which should not be in-use.
        for pt in self.get_policy_targets(context, {'id': pt_ids}):
            self.delete_policy_target(context, pt['id'])

        # REVISIT: The DB layer method called here manages its own
        # transaction, but ideally would be called within the same
        # transaction as the precommit calls above. Maybe the cleanup
        # of PRSs, proxy PTGs and PTs, which much be done outside any
        # transaction, could be moved before the initial transaction,
        # possibly within a loop to ensure no PTs concurrently become
        # in-use before proceding with the PTG deletion transaction.
        super(GroupPolicyPlugin, self).delete_policy_target_group(
            context, policy_target_group_id)

        try:
            self.policy_driver_manager.delete_policy_target_group_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_target_group_postcommit failed "
                          "for policy_target_group %s",
                          policy_target_group_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_target_group(self, context, policy_target_group_id,
                                fields=None):
        return self._get_resource(context, 'policy_target_group',
                                  policy_target_group_id,
                                  'PolicyTargetGroupContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_target_groups(self, context, filters=None, fields=None,
                                 sorts=None, limit=None, marker=None,
                                 page_reverse=False):
        return self._get_resources(
            context, 'policy_target_group', 'PolicyTargetGroupContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_application_policy_group(self, context,
                                        application_policy_group):
        self._ensure_tenant(
            context, application_policy_group['application_policy_group'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            pdm = self.policy_driver_manager
            result = super(GroupPolicyPlugin,
                           self).create_application_policy_group(
                               context, application_policy_group)
            self.extension_manager.process_create_application_policy_group(
                session, application_policy_group, result)
            self._validate_shared_create(self, context, result,
                                         'application_policy_group')
            policy_context = p_context.ApplicationPolicyGroupContext(
                self, context, result)
            pdm.create_application_policy_group_precommit(policy_context)

        try:
            pdm.create_application_policy_group_postcommit(policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_application_policy_group_postcommit "
                              "failed, deleting APG %s",
                              result['id'])
                self.delete_application_policy_group(context, result['id'])

        return self.get_application_policy_group(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_application_policy_group(self, context,
                                        application_policy_group_id,
                                        application_policy_group):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            pdm = self.policy_driver_manager
            original_application_policy_group = (
                self.get_application_policy_group(
                    context, application_policy_group_id))
            updated_application_policy_group = super(
                GroupPolicyPlugin, self).update_application_policy_group(
                    context, application_policy_group_id,
                    application_policy_group)

            self.extension_manager.process_update_application_policy_group(
                session, application_policy_group,
                updated_application_policy_group)
            self._validate_shared_update(
                self, context, original_application_policy_group,
                updated_application_policy_group, 'application_policy_group')
            policy_context = p_context.ApplicationPolicyGroupContext(
                self, context, updated_application_policy_group,
                original_application_policy_group=(
                    original_application_policy_group))
            pdm.update_application_policy_group_precommit(policy_context)

        pdm.update_application_policy_group_postcommit(policy_context)

        return self.get_application_policy_group(context,
                                                 application_policy_group_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_application_policy_group(self, context,
                                        application_policy_group_id):
        with db_api.CONTEXT_WRITER.using(context):
            pdm = self.policy_driver_manager
            application_policy_group = self.get_application_policy_group(
                context, application_policy_group_id)
            policy_context = p_context.ApplicationPolicyGroupContext(
                self, context, application_policy_group)
            pdm.delete_application_policy_group_precommit(policy_context)

            super(GroupPolicyPlugin, self).delete_application_policy_group(
                context, application_policy_group_id)

        try:
            pdm.delete_application_policy_group_postcommit(policy_context)
        except Exception:
            LOG.exception("delete_application_policy_group_postcommit "
                          "failed for application_policy_group %s",
                          application_policy_group_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_application_policy_group(self, context,
                                     application_policy_group_id, fields=None):
        return self._get_resource(context, 'application_policy_group',
                                  application_policy_group_id,
                                  'ApplicationPolicyGroupContext',
                                  fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_application_policy_groups(self, context, filters=None, fields=None,
                                      sorts=None, limit=None, marker=None,
                                      page_reverse=False):
        return self._get_resources(
            context, 'application_policy_group',
            'ApplicationPolicyGroupContext', filters=filters, fields=fields,
            sorts=sorts, limit=limit, marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_l2_policy(self, context, l2_policy):
        self._ensure_tenant(context, l2_policy['l2_policy'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_l2_policy(context, l2_policy)
            self.extension_manager.process_create_l2_policy(
                session, l2_policy, result)
            self._validate_shared_create(self, context, result, 'l2_policy')
            policy_context = p_context.L2PolicyContext(self, context, result)
            self.policy_driver_manager.create_l2_policy_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_l2_policy_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_l2_policy_postcommit "
                              "failed, deleting l2_policy %s",
                              result['id'])
                self.delete_l2_policy(context, result['id'])

        return self.get_l2_policy(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_l2_policy(self, context, l2_policy_id, l2_policy):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_l2_policy = self.get_l2_policy(context, l2_policy_id)
            updated_l2_policy = super(GroupPolicyPlugin,
                                      self).update_l2_policy(
                                          context, l2_policy_id, l2_policy)
            self.extension_manager.process_update_l2_policy(
                session, l2_policy, updated_l2_policy)
            self._validate_shared_update(self, context, original_l2_policy,
                                         updated_l2_policy, 'l2_policy')
            policy_context = p_context.L2PolicyContext(
                self, context, updated_l2_policy,
                original_l2_policy=original_l2_policy)
            self.policy_driver_manager.update_l2_policy_precommit(
                policy_context)

        self.policy_driver_manager.update_l2_policy_postcommit(
            policy_context)

        return self.get_l2_policy(context, l2_policy_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_l2_policy(self, context, l2_policy_id):
        with db_api.CONTEXT_WRITER.using(context):
            l2_policy = self.get_l2_policy(context, l2_policy_id)
            policy_context = p_context.L2PolicyContext(self, context,
                                                       l2_policy)
            self.policy_driver_manager.delete_l2_policy_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_l2_policy(context,
                                                            l2_policy_id)

        try:
            self.policy_driver_manager.delete_l2_policy_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_l2_policy_postcommit failed "
                          "for l2_policy %s", l2_policy_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_l2_policy(self, context, l2_policy_id, fields=None):
        return self._get_resource(context, 'l2_policy',
                                  l2_policy_id,
                                  'L2PolicyContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_l2_policies(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        return self._get_resources(
            context, 'l2_policy', 'L2PolicyContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_network_service_policy(self, context, network_service_policy):
        self._ensure_tenant(
            context, network_service_policy['network_service_policy'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_network_service_policy(
                               context, network_service_policy)
            self.extension_manager.process_create_network_service_policy(
                session, network_service_policy, result)
            self._validate_shared_create(self, context, result,
                                         'network_service_policy')
            policy_context = p_context.NetworkServicePolicyContext(
                self, context, result)
            pdm = self.policy_driver_manager
            pdm.create_network_service_policy_precommit(
                policy_context)

        try:
            pdm.create_network_service_policy_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "create_network_service_policy_postcommit "
                    "failed, deleting network_service_policy %s",
                    result['id'])
                self.delete_network_service_policy(context, result['id'])

        return self.get_network_service_policy(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_network_service_policy(self, context, network_service_policy_id,
                                      network_service_policy):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_network_service_policy = super(
                GroupPolicyPlugin, self).get_network_service_policy(
                    context, network_service_policy_id)
            updated_network_service_policy = super(
                GroupPolicyPlugin, self).update_network_service_policy(
                    context, network_service_policy_id, network_service_policy)
            self.extension_manager.process_update_network_service_policy(
                session, network_service_policy,
                updated_network_service_policy)
            self._validate_shared_update(
                self, context, original_network_service_policy,
                updated_network_service_policy, 'network_service_policy')
            policy_context = p_context.NetworkServicePolicyContext(
                self, context, updated_network_service_policy,
                original_network_service_policy=(
                    original_network_service_policy))
            self.policy_driver_manager.update_network_service_policy_precommit(
                policy_context)

        self.policy_driver_manager.update_network_service_policy_postcommit(
            policy_context)
        return self.get_network_service_policy(context,
                                               network_service_policy_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_network_service_policy(
        self, context, network_service_policy_id):
        with db_api.CONTEXT_WRITER.using(context):
            network_service_policy = self.get_network_service_policy(
                context, network_service_policy_id)
            policy_context = p_context.NetworkServicePolicyContext(
                self, context, network_service_policy)
            self.policy_driver_manager.delete_network_service_policy_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_network_service_policy(
                context, network_service_policy_id)

        try:
            pdm = self.policy_driver_manager
            pdm.delete_network_service_policy_postcommit(policy_context)
        except Exception:
            LOG.exception(
                "delete_network_service_policy_postcommit failed "
                "for network_service_policy %s", network_service_policy_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_network_service_policy(self, context, network_service_policy_id,
                                   fields=None):
        return self._get_resource(context, 'network_service_policy',
                                  network_service_policy_id,
                                  'NetworkServicePolicyContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_network_service_policies(self, context, filters=None, fields=None,
                                     sorts=None, limit=None, marker=None,
                                     page_reverse=False):
        return self._get_resources(
            context, 'network_service_policy', 'NetworkServicePolicyContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_l3_policy(self, context, l3_policy):
        self._ensure_tenant(context, l3_policy['l3_policy'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_l3_policy(context, l3_policy)
            self.extension_manager.process_create_l3_policy(
                session, l3_policy, result)
            self._validate_shared_create(self, context, result, 'l3_policy')
            self._validate_l3p_es(context, result)
            policy_context = p_context.L3PolicyContext(self, context,
                                                       result)
            self.policy_driver_manager.create_l3_policy_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_l3_policy_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_l3_policy_postcommit "
                              "failed, deleting l3_policy %s",
                              result['id'])
                self.delete_l3_policy(context, result['id'])

        return self.get_l3_policy(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_l3_policy(self, context, l3_policy_id, l3_policy):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_l3_policy = self.get_l3_policy(context, l3_policy_id)
            updated_l3_policy = super(
                GroupPolicyPlugin, self).update_l3_policy(
                    context, l3_policy_id, l3_policy)
            self.extension_manager.process_update_l3_policy(
                session, l3_policy, updated_l3_policy)
            self._validate_shared_update(self, context, original_l3_policy,
                                         updated_l3_policy, 'l3_policy')
            self._validate_l3p_es(context, updated_l3_policy,
                                  original_l3_policy)
            policy_context = p_context.L3PolicyContext(
                self, context, updated_l3_policy,
                original_l3_policy=original_l3_policy)
            self.policy_driver_manager.update_l3_policy_precommit(
                policy_context)

        self.policy_driver_manager.update_l3_policy_postcommit(
            policy_context)
        return self.get_l3_policy(context, l3_policy_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_l3_policy(self, context, l3_policy_id, check_unused=False):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            if (check_unused and
                (session.query(group_policy_mapping_db.L2PolicyMapping).
                 filter_by(l3_policy_id=l3_policy_id).count())):
                return False
            l3_policy = self.get_l3_policy(context, l3_policy_id)
            policy_context = p_context.L3PolicyContext(self, context,
                                                       l3_policy)
            self.policy_driver_manager.delete_l3_policy_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_l3_policy(context,
                                                            l3_policy_id)

        try:
            self.policy_driver_manager.delete_l3_policy_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_l3_policy_postcommit failed "
                          "for l3_policy %s", l3_policy_id)
        return True

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_l3_policy(self, context, l3_policy_id, fields=None):
        return self._get_resource(context, 'l3_policy',
                                  l3_policy_id,
                                  'L3PolicyContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_l3_policies(self, context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
        return self._get_resources(
            context, 'l3_policy', 'L3PolicyContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_classifier(self, context, policy_classifier):
        self._ensure_tenant(context, policy_classifier['policy_classifier'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(
                GroupPolicyPlugin, self).create_policy_classifier(
                    context, policy_classifier)
            self.extension_manager.process_create_policy_classifier(
                session, policy_classifier, result)
            self._validate_shared_create(
                self, context, result, 'policy_classifier')
            policy_context = p_context.PolicyClassifierContext(self, context,
                                                               result)
            self.policy_driver_manager.create_policy_classifier_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_classifier_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "policy_driver_manager.create_policy_classifier_postcommit"
                    " failed, deleting policy_classifier %s", result['id'])
                self.delete_policy_classifier(context, result['id'])

        return self.get_policy_classifier(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_classifier(self, context, id, policy_classifier):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_policy_classifier = super(
                GroupPolicyPlugin, self).get_policy_classifier(context, id)
            updated_policy_classifier = super(
                GroupPolicyPlugin, self).update_policy_classifier(
                    context, id, policy_classifier)
            self.extension_manager.process_update_policy_classifier(
                session, policy_classifier, updated_policy_classifier)
            self._validate_shared_update(
                self, context, original_policy_classifier,
                updated_policy_classifier, 'policy_classifier')
            policy_context = p_context.PolicyClassifierContext(
                self, context, updated_policy_classifier,
                original_policy_classifier=original_policy_classifier)
            self.policy_driver_manager.update_policy_classifier_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_classifier_postcommit(
            policy_context)
        return self.get_policy_classifier(context, id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_classifier(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_classifier = self.get_policy_classifier(context, id)
            policy_context = p_context.PolicyClassifierContext(
                self, context, policy_classifier)
            self.policy_driver_manager.delete_policy_classifier_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_policy_classifier(
                context, id)

        try:
            self.policy_driver_manager.delete_policy_classifier_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_classifier_postcommit failed "
                          "for policy_classifier %s", id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_classifier(self, context, policy_classifier_id,
                              fields=None):
        return self._get_resource(context, 'policy_classifier',
                                  policy_classifier_id,
                                  'PolicyClassifierContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_classifiers(self, context, filters=None, fields=None,
                               sorts=None, limit=None, marker=None,
                               page_reverse=False):
        return self._get_resources(
            context, 'policy_classifier', 'PolicyClassifierContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_action(self, context, policy_action):
        self._ensure_tenant(context, policy_action['policy_action'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_policy_action(context, policy_action)
            self.extension_manager.process_create_policy_action(
                session, policy_action, result)
            self._validate_shared_create(self, context, result,
                                         'policy_action')
            self._validate_action_value(context, result)
            policy_context = p_context.PolicyActionContext(self, context,
                                                           result)
            self.policy_driver_manager.create_policy_action_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_action_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "policy_driver_manager.create_policy_action_postcommit "
                    "failed, deleting policy_action %s", result['id'])
                self.delete_policy_action(context, result['id'])

        return self.get_policy_action(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_action(self, context, id, policy_action):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_policy_action = super(
                GroupPolicyPlugin, self).get_policy_action(context, id)
            updated_policy_action = super(
                GroupPolicyPlugin, self).update_policy_action(context, id,
                                                              policy_action)
            self.extension_manager.process_update_policy_action(
                session, policy_action, updated_policy_action)
            self._validate_shared_update(self, context, original_policy_action,
                                         updated_policy_action,
                                         'policy_action')
            self._validate_action_value(context, updated_policy_action)
            policy_context = p_context.PolicyActionContext(
                self, context, updated_policy_action,
                original_policy_action=original_policy_action)
            self.policy_driver_manager.update_policy_action_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_action_postcommit(
            policy_context)
        return self.get_policy_action(context, id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_action(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_action = self.get_policy_action(context, id)
            policy_context = p_context.PolicyActionContext(self, context,
                                                           policy_action)
            self.policy_driver_manager.delete_policy_action_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_policy_action(context, id)

        try:
            self.policy_driver_manager.delete_policy_action_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_action_postcommit failed "
                          "for policy_action %s", id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_action(self, context, policy_action_id, fields=None):
        return self._get_resource(context, 'policy_action',
                                  policy_action_id,
                                  'PolicyActionContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_actions(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        return self._get_resources(
            context, 'policy_action', 'PolicyActionContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_rule(self, context, policy_rule):
        self._ensure_tenant(context, policy_rule['policy_rule'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(
                GroupPolicyPlugin, self).create_policy_rule(
                    context, policy_rule)
            self.extension_manager.process_create_policy_rule(
                session, policy_rule, result)
            self._validate_shared_create(self, context, result, 'policy_rule')
            policy_context = p_context.PolicyRuleContext(self, context,
                                                         result)
            self.policy_driver_manager.create_policy_rule_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_rule_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "policy_driver_manager.create_policy_rule_postcommit"
                    " failed, deleting policy_rule %s", result['id'])
                self.delete_policy_rule(context, result['id'])

        return self.get_policy_rule(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_rule(self, context, id, policy_rule):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_policy_rule = super(
                GroupPolicyPlugin, self).get_policy_rule(context, id)
            updated_policy_rule = super(
                GroupPolicyPlugin, self).update_policy_rule(
                    context, id, policy_rule)
            self.extension_manager.process_update_policy_rule(
                session, policy_rule, updated_policy_rule)
            self._validate_shared_update(self, context, original_policy_rule,
                                         updated_policy_rule, 'policy_rule')
            policy_context = p_context.PolicyRuleContext(
                self, context, updated_policy_rule,
                original_policy_rule=original_policy_rule)
            self.policy_driver_manager.update_policy_rule_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_rule_postcommit(
            policy_context)
        return self.get_policy_rule(context, id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_rule(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_rule = self.get_policy_rule(context, id)
            policy_context = p_context.PolicyRuleContext(self, context,
                                                         policy_rule)
            self.policy_driver_manager.delete_policy_rule_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_policy_rule(
                context, id)

        try:
            self.policy_driver_manager.delete_policy_rule_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_rule_postcommit failed "
                          "for policy_rule %s", id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_rule(self, context, policy_rule_id, fields=None):
        return self._get_resource(context, 'policy_rule',
                                  policy_rule_id,
                                  'PolicyRuleContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_rules(self, context, filters=None, fields=None,
                         sorts=None, limit=None, marker=None,
                         page_reverse=False):
        return self._get_resources(
            context, 'policy_rule', 'PolicyRuleContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_policy_rule_set(self, context, policy_rule_set):
        self._ensure_tenant(context, policy_rule_set['policy_rule_set'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_policy_rule_set(
                               context, policy_rule_set)
            self.extension_manager.process_create_policy_rule_set(
                session, policy_rule_set, result)
            self._validate_shared_create(
                self, context, result, 'policy_rule_set')
            policy_context = p_context.PolicyRuleSetContext(
                self, context, result)
            self.policy_driver_manager.create_policy_rule_set_precommit(
                policy_context)

        try:
            self.policy_driver_manager.create_policy_rule_set_postcommit(
                policy_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "policy_driver_manager.create_policy_rule_set_postcommit "
                    "failed, deleting policy_rule_set %s", result['id'])
                self.delete_policy_rule_set(context, result['id'])

        return self.get_policy_rule_set(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_policy_rule_set(self, context, id, policy_rule_set):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_policy_rule_set = super(
                GroupPolicyPlugin, self).get_policy_rule_set(context, id)
            updated_policy_rule_set = super(
                GroupPolicyPlugin, self).update_policy_rule_set(
                    context, id, policy_rule_set)
            self.extension_manager.process_update_policy_rule_set(
                session, policy_rule_set, updated_policy_rule_set)
            self._validate_shared_update(
                self, context, original_policy_rule_set,
                updated_policy_rule_set, 'policy_rule_set')
            policy_context = p_context.PolicyRuleSetContext(
                self, context, updated_policy_rule_set,
                original_policy_rule_set=original_policy_rule_set)
            self.policy_driver_manager.update_policy_rule_set_precommit(
                policy_context)

        self.policy_driver_manager.update_policy_rule_set_postcommit(
            policy_context)
        return self.get_policy_rule_set(context, id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_policy_rule_set(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            policy_rule_set = self.get_policy_rule_set(context, id)
            policy_context = p_context.PolicyRuleSetContext(
                self, context, policy_rule_set)
            self.policy_driver_manager.delete_policy_rule_set_precommit(
                policy_context)
            super(GroupPolicyPlugin, self).delete_policy_rule_set(context, id)

        try:
            self.policy_driver_manager.delete_policy_rule_set_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_policy_rule_set_postcommit failed "
                          "for policy_rule_set %s", id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_rule_set(self, context, policy_rule_set_id, fields=None):
        return self._get_resource(context, 'policy_rule_set',
                                  policy_rule_set_id,
                                  'PolicyRuleSetContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_policy_rule_sets(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        return self._get_resources(
            context, 'policy_rule_set', 'PolicyRuleSetContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_external_segment(self, context, external_segment):
        self._ensure_tenant(context, external_segment['external_segment'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_external_segment(context,
                                                         external_segment)
            self.extension_manager.process_create_external_segment(
                session, external_segment, result)
            self._validate_shared_create(self, context, result,
                                         'external_segment')
            policy_context = p_context.ExternalSegmentContext(
                self, context, result)
            (self.policy_driver_manager.
             create_external_segment_precommit(policy_context))
            # Validate the routes after the drivers had the chance to fill
            # the cidr field.
            self._validate_routes(context, result)

        try:
            (self.policy_driver_manager.
             create_external_segment_postcommit(policy_context))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_external_segment_postcommit "
                              "failed, deleting external_segment "
                              "%s", result['id'])
                self.delete_external_segment(context, result['id'])

        return self.get_external_segment(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_external_segment(self, context, external_segment_id,
                                external_segment):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_external_segment = super(
                GroupPolicyPlugin, self).get_external_segment(
                    context, external_segment_id)
            updated_external_segment = super(
                GroupPolicyPlugin, self).update_external_segment(
                    context, external_segment_id,
                    external_segment)
            self.extension_manager.process_update_external_segment(
                session, external_segment, updated_external_segment)
            self._validate_shared_update(
                self, context, original_external_segment,
                updated_external_segment, 'external_segment')
            self._validate_routes(context, updated_external_segment,
                                  original_external_segment)
            # TODO(ivar): Validate Routes' GW in es subnet
            policy_context = p_context.ExternalSegmentContext(
                self, context, updated_external_segment,
                original_external_segment)
            (self.policy_driver_manager.
             update_external_segment_precommit(policy_context))

        self.policy_driver_manager.update_external_segment_postcommit(
            policy_context)
        return self.get_external_segment(context, external_segment_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_external_segment(self, context, external_segment_id):
        with db_api.CONTEXT_WRITER.using(context):
            es = self.get_external_segment(context, external_segment_id)
            if es['l3_policies'] or es['nat_pools'] or es['external_policies']:
                raise gpex.ExternalSegmentInUse(es_id=es['id'])
            policy_context = p_context.ExternalSegmentContext(
                self, context, es)
            (self.policy_driver_manager.
             delete_external_segment_precommit(policy_context))
            super(GroupPolicyPlugin, self).delete_external_segment(
                context, external_segment_id)

        try:
            (self.policy_driver_manager.
             delete_external_segment_postcommit(policy_context))
        except Exception:
            LOG.exception("delete_external_segment_postcommit failed "
                          "for external_segment %s",
                          external_segment_id)
        return True

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_external_segment(self, context, external_segment_id, fields=None):
        return self._get_resource(context, 'external_segment',
                                  external_segment_id,
                                  'ExternalSegmentContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_external_segments(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        return self._get_resources(
            context, 'external_segment', 'ExternalSegmentContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_external_policy(self, context, external_policy):
        self._ensure_tenant(context, external_policy['external_policy'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin,
                           self).create_external_policy(
                               context, external_policy)
            self.extension_manager.process_create_external_policy(
                session, external_policy, result)
            self._validate_shared_create(self, context, result,
                                         'external_policy')
            policy_context = p_context.ExternalPolicyContext(
                self, context, result)
            (self.policy_driver_manager.
             create_external_policy_precommit(policy_context))

        try:
            (self.policy_driver_manager.
             create_external_policy_postcommit(policy_context))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("create_external_policy_postcommit "
                              "failed, deleting external_policy "
                              "%s", result['id'])
                self.delete_external_policy(context, result['id'])

        return self.get_external_policy(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_external_policy(self, context, external_policy_id,
                               external_policy):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_external_policy = super(
                GroupPolicyPlugin, self).get_external_policy(
                    context, external_policy_id)
            updated_external_policy = super(
                GroupPolicyPlugin, self).update_external_policy(
                    context, external_policy_id,
                    external_policy)
            self.extension_manager.process_update_external_policy(
                session, external_policy, updated_external_policy)
            self._validate_shared_update(
                self, context, original_external_policy,
                updated_external_policy, 'external_policy')
            policy_context = p_context.ExternalPolicyContext(
                self, context, updated_external_policy,
                original_external_policy)
            (self.policy_driver_manager.
             update_external_policy_precommit(policy_context))

        self.policy_driver_manager.update_external_policy_postcommit(
            policy_context)
        return self.get_external_policy(context, external_policy_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_external_policy(self, context, external_policy_id,
                               check_unused=False):
        with db_api.CONTEXT_WRITER.using(context):
            es = self.get_external_policy(context, external_policy_id)
            policy_context = p_context.ExternalPolicyContext(
                self, context, es)
            (self.policy_driver_manager.
             delete_external_policy_precommit(policy_context))
            super(GroupPolicyPlugin, self).delete_external_policy(
                context, external_policy_id)

        try:
            self.policy_driver_manager.delete_external_policy_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_external_policy_postcommit failed "
                          "for external_policy %s", external_policy_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_external_policy(self, context, external_policy_id, fields=None):
        return self._get_resource(context, 'external_policy',
                                  external_policy_id,
                                  'ExternalPolicyContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_external_policies(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        return self._get_resources(
            context, 'external_policy', 'ExternalPolicyContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def create_nat_pool(self, context, nat_pool):
        self._ensure_tenant(context, nat_pool['nat_pool'])
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            result = super(GroupPolicyPlugin, self).create_nat_pool(
                context, nat_pool)
            self.extension_manager.process_create_nat_pool(session, nat_pool,
                                                           result)
            self._validate_shared_create(self, context, result, 'nat_pool')
            policy_context = p_context.NatPoolContext(self, context, result)
            (self.policy_driver_manager.
             create_nat_pool_precommit(policy_context))

        try:
            (self.policy_driver_manager.
             create_nat_pool_postcommit(policy_context))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception(
                    "create_nat_pool_postcommit failed, deleting "
                    "nat_pool %s", result['id'])
                self.delete_nat_pool(context, result['id'])

        return self.get_nat_pool(context, result['id'])

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def update_nat_pool(self, context, nat_pool_id, nat_pool):
        with db_api.CONTEXT_WRITER.using(context):
            session = context.session
            original_nat_pool = super(
                GroupPolicyPlugin, self).get_nat_pool(context, nat_pool_id)
            updated_nat_pool = super(
                GroupPolicyPlugin, self).update_nat_pool(context, nat_pool_id,
                                                         nat_pool)
            self.extension_manager.process_update_nat_pool(
                session, nat_pool, updated_nat_pool)
            self._validate_shared_update(self, context, original_nat_pool,
                                         updated_nat_pool, 'nat_pool')
            policy_context = p_context.NatPoolContext(
                self, context, updated_nat_pool, original_nat_pool)
            (self.policy_driver_manager.
             update_nat_pool_precommit(policy_context))

        self.policy_driver_manager.update_nat_pool_postcommit(policy_context)
        return self.get_nat_pool(context, nat_pool_id)

    @log.log_method_call
    @n_utils.transaction_guard
    @db_api.retry_if_session_inactive()
    def delete_nat_pool(self, context, nat_pool_id, check_unused=False):
        with db_api.CONTEXT_WRITER.using(context):
            es = self.get_nat_pool(context, nat_pool_id)
            policy_context = p_context.NatPoolContext(self, context, es)
            (self.policy_driver_manager.delete_nat_pool_precommit(
                policy_context))
            super(GroupPolicyPlugin, self).delete_nat_pool(context,
                                                           nat_pool_id)

        try:
            self.policy_driver_manager.delete_nat_pool_postcommit(
                policy_context)
        except Exception:
            LOG.exception("delete_nat_pool_postcommit failed "
                          "for nat_pool %s",
                          nat_pool_id)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_nat_pool(self, context, nat_pool_id, fields=None):
        return self._get_resource(context, 'nat_pool',
                                  nat_pool_id,
                                  'NatPoolContext', fields=fields)

    @log.log_method_call
    @db_api.retry_if_session_inactive()
    def get_nat_pools(self, context, filters=None, fields=None,
                      sorts=None, limit=None, marker=None,
                      page_reverse=False):
        return self._get_resources(
            context, 'nat_pool', 'NatPoolContext',
            filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    def _is_port_bound(self, port_id):
        # REVISIT(ivar): This operation shouldn't be done within a DB lock
        # once we refactor the server.
        not_bound = [portbindings.VIF_TYPE_UNBOUND,
                     portbindings.VIF_TYPE_BINDING_FAILED]
        context = n_ctx.get_admin_context()
        port = directory.get_plugin().get_port(context, port_id)
        return (port.get('binding:vif_type') not in not_bound) and port.get(
            'binding:host_id') and (port['device_owner'] or port['device_id'])

    def _is_service_target(self, context, pt_id):
        return bool(ncp_model.get_service_targets_count(
            context.session, pt_id))

    def _ensure_tenant(self, context, resource):
        # TODO(Sumit): This check is ideally not required, but a bunch of UTs
        # are not setup correctly to populate the tenant_id, hence we
        # temporarily need to perform this check. This will go with the fix
        # for the deprecated get_tenant_id_for_create method.
        if 'tenant_id' in resource:
            tenant_id = resource['tenant_id']
            self.policy_driver_manager.ensure_tenant(context, tenant_id)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from aim.api import resource as aim_res
from aim import exceptions as aim_exc
from neutron.api import extensions
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import directory
from oslo_log import log
from oslo_utils import excutils

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron import extensions as extensions_pkg
from gbpservice.neutron.extensions import cisco_apic
from gbpservice.neutron.plugins.ml2plus import driver_api as api_plus
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import (
    extension_db as extn_db)
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import db

LOG = log.getLogger(__name__)


class ApicExtensionDriver(api_plus.ExtensionDriver,
                          db.DbMixin,
                          extn_db.ExtensionDbMixin):

    def __init__(self):
        LOG.info("APIC AIM ED __init__")
        self._mechanism_driver = None

    def initialize(self):
        LOG.info("APIC AIM ED initializing")
        extensions.append_api_extensions_path(extensions_pkg.__path__)

    @property
    def _md(self):
        if not self._mechanism_driver:
            # REVISIT(rkukura): It might be safer to search the MDs by
            # class rather than index by name, or to use a class
            # variable to find the instance.
            plugin = directory.get_plugin()
            mech_mgr = plugin.mechanism_manager
            self._mechanism_driver = mech_mgr.mech_drivers['apic_aim'].obj
        return self._mechanism_driver

    @property
    def extension_alias(self):
        return "cisco-apic"

    def extend_port_dict(self, session, base_model, result):
        try:
            self._md.extend_port_dict(session, base_model, result)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_port_dict got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_port_dict failed")

    def extend_port_dict_bulk(self, session, results):
        try:
            self._md.extend_port_dict_bulk(session, results)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_port_dict_bulk got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_port_dict_bulk failed")

    def process_create_port(self, plugin_context, data, result):
        res_dict = {cisco_apic.ERSPAN_CONFIG:
                    data.get(cisco_apic.ERSPAN_CONFIG, [])}
        self.set_port_extn_db(plugin_context.session, result['id'],
                              res_dict)
        result.update(res_dict)

    def process_update_port(self, plugin_context, data, result):
        if cisco_apic.ERSPAN_CONFIG not in data:
            return
        res_dict = {cisco_apic.ERSPAN_CONFIG: data[cisco_apic.ERSPAN_CONFIG]}
        self.set_port_extn_db(plugin_context.session, result['id'],
                              res_dict)
        result.update(res_dict)

    def extend_network_dict(self, session, base_model, result):
        try:
            self._md.extend_network_dict(session, base_model, result)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_network_dict got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_network_dict failed")

    def extend_network_dict_bulk(self, session, results):
        try:
            self._md.extend_network_dict_bulk(session, results)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_network_dict got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_network_dict failed")

    def validate_bgp_params(self, data, result=None):
        if result:
            is_svi = result.get(cisco_apic.SVI)
        else:
            is_svi = data.get(cisco_apic.SVI, False)
        is_bgp_enabled = data.get(cisco_apic.BGP, False)
        bgp_type = data.get(cisco_apic.BGP_TYPE, "default_export")
        asn = data.get(cisco_apic.BGP_ASN, "0")
        if not is_svi and (is_bgp_enabled or (bgp_type != "default_export") or
                           (asn != "0")):
            raise n_exc.InvalidInput(error_message="Network has to be created"
                                     " as svi type(--apic:svi True) to enable"
                                     " BGP or to set BGP parameters")

    def process_create_network(self, plugin_context, data, result):
        is_svi = data.get(cisco_apic.SVI, False)
        is_bgp_enabled = data.get(cisco_apic.BGP, False)
        bgp_type = data.get(cisco_apic.BGP_TYPE, "default_export")
        asn = data.get(cisco_apic.BGP_ASN, "0")
        use_multi_ext_nets = data.get(cisco_apic.MULTI_EXT_NETS, False)
        self.validate_bgp_params(data)
        res_dict = {cisco_apic.SVI: is_svi,
                    cisco_apic.BGP: is_bgp_enabled,
                    cisco_apic.BGP_TYPE: bgp_type,
                    cisco_apic.BGP_ASN: asn,
                    cisco_apic.NESTED_DOMAIN_NAME:
                    data.get(cisco_apic.NESTED_DOMAIN_NAME),
                    cisco_apic.NESTED_DOMAIN_TYPE:
                    data.get(cisco_apic.NESTED_DOMAIN_TYPE),
                    cisco_apic.NESTED_DOMAIN_INFRA_VLAN:
                    data.get(cisco_apic.NESTED_DOMAIN_INFRA_VLAN),
                    cisco_apic.NESTED_DOMAIN_SERVICE_VLAN:
                    data.get(cisco_apic.NESTED_DOMAIN_SERVICE_VLAN),
                    cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN:
                    data.get(cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN),
                    cisco_apic.EXTRA_PROVIDED_CONTRACTS:
                    data.get(cisco_apic.EXTRA_PROVIDED_CONTRACTS),
                    cisco_apic.EXTRA_CONSUMED_CONTRACTS:
                    data.get(cisco_apic.EXTRA_CONSUMED_CONTRACTS),
                    cisco_apic.EPG_CONTRACT_MASTERS:
                    data.get(cisco_apic.EPG_CONTRACT_MASTERS),
                    cisco_apic.POLICY_ENFORCEMENT_PREF:
                    data.get(cisco_apic.POLICY_ENFORCEMENT_PREF, "unenforced"),
                    cisco_apic.NO_NAT_CIDRS:
                    data.get(cisco_apic.NO_NAT_CIDRS),
                    cisco_apic.MULTI_EXT_NETS: use_multi_ext_nets,
                    }
        if cisco_apic.VLANS_LIST in (data.get(
                cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS) or {}):
            res_dict.update({cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS:
                data.get(cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS)[
                    cisco_apic.VLANS_LIST]})
        self.set_network_extn_db(plugin_context.session, result['id'],
                                 res_dict)
        result.update(res_dict)

        if (data.get(cisco_apic.DIST_NAMES) and
            data[cisco_apic.DIST_NAMES].get(cisco_apic.EXTERNAL_NETWORK)):
            dn = data[cisco_apic.DIST_NAMES][cisco_apic.EXTERNAL_NETWORK]
            try:
                aim_res.ExternalNetwork.from_dn(dn)
            except aim_exc.InvalidDNForAciResource:
                raise n_exc.InvalidInput(
                    error_message=('%s is not valid ExternalNetwork DN' % dn))
            if is_svi:
                res_dict = {cisco_apic.EXTERNAL_NETWORK: dn}
            else:
                res_dict = {cisco_apic.EXTERNAL_NETWORK: dn,
                            cisco_apic.NAT_TYPE:
                            data.get(cisco_apic.NAT_TYPE, 'distributed'),
                            cisco_apic.EXTERNAL_CIDRS:
                            data.get(
                                cisco_apic.EXTERNAL_CIDRS, ['0.0.0.0/0'])}
            self.set_network_extn_db(plugin_context.session, result['id'],
                                     res_dict)
            result.setdefault(cisco_apic.DIST_NAMES, {})[
                    cisco_apic.EXTERNAL_NETWORK] = res_dict.pop(
                        cisco_apic.EXTERNAL_NETWORK)
            result.update(res_dict)
        if (data.get(cisco_apic.DIST_NAMES) and
            data[cisco_apic.DIST_NAMES].get(cisco_apic.BD)):
            dn = data[cisco_apic.DIST_NAMES][cisco_apic.BD]
            try:
                aim_res.BridgeDomain.from_dn(dn)
            except aim_exc.InvalidDNForAciResource:
                raise n_exc.InvalidInput(
                    error_message=('%s is not valid BridgeDomain DN' % dn))
            res_dict = {cisco_apic.BD: dn}
            self.set_network_extn_db(plugin_context.session, result['id'],
                                     res_dict)
            result.setdefault(cisco_apic.DIST_NAMES, {})[
                    cisco_apic.BD] = res_dict.pop(
                        cisco_apic.BD)
            result.update(res_dict)

    def process_update_network(self, plugin_context, data, result):
        # Extension attributes that could be updated.
        update_attrs = [
                cisco_apic.EXTERNAL_CIDRS, cisco_apic.BGP, cisco_apic.BGP_TYPE,
                cisco_apic.BGP_ASN,
                cisco_apic.NESTED_DOMAIN_NAME, cisco_apic.NESTED_DOMAIN_TYPE,
                cisco_apic.NESTED_DOMAIN_INFRA_VLAN,
                cisco_apic.NESTED_DOMAIN_SERVICE_VLAN,
                cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN,
                cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS,
                cisco_apic.EXTRA_PROVIDED_CONTRACTS,
                cisco_apic.EXTRA_CONSUMED_CONTRACTS,
                cisco_apic.EPG_CONTRACT_MASTERS,
                cisco_apic.POLICY_ENFORCEMENT_PREF,
                cisco_apic.NO_NAT_CIDRS,
                cisco_apic.MULTI_EXT_NETS]
        if not (set(update_attrs) & set(data.keys())):
            return

        res_dict = {}
        if result.get(cisco_apic.DIST_NAMES, {}).get(
            cisco_apic.EXTERNAL_NETWORK):
            if cisco_apic.EXTERNAL_CIDRS in data:
                res_dict.update({cisco_apic.EXTERNAL_CIDRS:
                    data[cisco_apic.EXTERNAL_CIDRS]})
        self.validate_bgp_params(data, result)

        ext_keys = [cisco_apic.BGP, cisco_apic.BGP_TYPE, cisco_apic.BGP_ASN,
                cisco_apic.NESTED_DOMAIN_NAME, cisco_apic.NESTED_DOMAIN_TYPE,
                cisco_apic.NESTED_DOMAIN_INFRA_VLAN,
                cisco_apic.NESTED_DOMAIN_SERVICE_VLAN,
                cisco_apic.NESTED_DOMAIN_NODE_NETWORK_VLAN,
                cisco_apic.EXTRA_PROVIDED_CONTRACTS,
                cisco_apic.EXTRA_CONSUMED_CONTRACTS,
                cisco_apic.EPG_CONTRACT_MASTERS,
                cisco_apic.POLICY_ENFORCEMENT_PREF,
                cisco_apic.NO_NAT_CIDRS,
                cisco_apic.MULTI_EXT_NETS]
        for e_k in ext_keys:
            if e_k in data:
                res_dict.update({e_k: data[e_k]})

        if cisco_apic.VLANS_LIST in (data.get(
                cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS) or {}):
            res_dict.update({cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS:
                data.get(cisco_apic.NESTED_DOMAIN_ALLOWED_VLANS)[
                    cisco_apic.VLANS_LIST]})

        if res_dict:
            self.set_network_extn_db(plugin_context.session, result['id'],
                                     res_dict)
            result.update(res_dict)

    def extend_subnet_dict(self, session, base_model, result):
        try:
            self._md.extend_subnet_dict(session, base_model, result)
            res_dict = self.get_subnet_extn_db(session, result['id'])
            result[cisco_apic.SNAT_HOST_POOL] = (
                res_dict.get(cisco_apic.SNAT_HOST_POOL, False))
            result[cisco_apic.ACTIVE_ACTIVE_AAP] = (
                res_dict.get(cisco_apic.ACTIVE_ACTIVE_AAP, False))
            result[cisco_apic.SNAT_SUBNET_ONLY] = (
                res_dict.get(cisco_apic.SNAT_SUBNET_ONLY, False))
            result[cisco_apic.EPG_SUBNET] = (
                res_dict.get(cisco_apic.EPG_SUBNET, False))
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_subnet_dict got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_subnet_dict failed")

    def extend_subnet_dict_bulk(self, session, results):
        try:
            self._md.extend_subnet_dict_bulk(session, results)
            for result, subnet_db in results:
                res_dict = self.get_subnet_extn_db(session, subnet_db['id'])
                result[cisco_apic.SNAT_HOST_POOL] = (
                    res_dict.get(cisco_apic.SNAT_HOST_POOL, False))
                result[cisco_apic.ACTIVE_ACTIVE_AAP] = (
                    res_dict.get(cisco_apic.ACTIVE_ACTIVE_AAP, False))
                result[cisco_apic.SNAT_SUBNET_ONLY] = (
                    res_dict.get(cisco_apic.SNAT_SUBNET_ONLY, False))
                result[cisco_apic.EPG_SUBNET] = (
                    res_dict.get(cisco_apic.EPG_SUBNET, False))
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_subnet_dict_bulk got retriable "
                              "exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_subnet_dict_bulk failed")

    def process_create_subnet(self, plugin_context, data, result):
        res_dict = {cisco_apic.SNAT_HOST_POOL:
                    data.get(cisco_apic.SNAT_HOST_POOL, False),
                    cisco_apic.ACTIVE_ACTIVE_AAP:
                    data.get(cisco_apic.ACTIVE_ACTIVE_AAP, False),
                    cisco_apic.SNAT_SUBNET_ONLY:
                    data.get(cisco_apic.SNAT_SUBNET_ONLY, False),
                    cisco_apic.EPG_SUBNET:
                    data.get(cisco_apic.EPG_SUBNET, False)}
        self.set_subnet_extn_db(plugin_context.session, result['id'],
                                res_dict)
        result.update(res_dict)

    def process_update_subnet(self, plugin_context, data, result):
        if (cisco_apic.SNAT_HOST_POOL not in data and
                cisco_apic.SNAT_SUBNET_ONLY not in data):
            return

        res_dict = {}
        if cisco_apic.SNAT_HOST_POOL in data:
            res_dict.update({cisco_apic.SNAT_HOST_POOL:
                             data[cisco_apic.SNAT_HOST_POOL]})

        if cisco_apic.SNAT_SUBNET_ONLY in data:
            res_dict.update({cisco_apic.SNAT_SUBNET_ONLY:
                             data[cisco_apic.SNAT_SUBNET_ONLY]})

        self.set_subnet_extn_db(plugin_context.session, result['id'],
                                res_dict)
        result.update(res_dict)

    def extend_address_scope_dict(self, session, base_model, result):
        try:
            self._md.extend_address_scope_dict(session, base_model, result)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                if db_api.is_retriable(e):
                    LOG.debug("APIC AIM extend_address_scope_dict got "
                              "retriable exception: %s", type(e))
                else:
                    LOG.exception("APIC AIM extend_address_scope_dict failed")

    def process_create_address_scope(self, plugin_context, data, result):
        if (data.get(cisco_apic.DIST_NAMES) and
            data[cisco_apic.DIST_NAMES].get(cisco_apic.VRF)):
            dn = data[cisco_apic.DIST_NAMES][cisco_apic.VRF]
            try:
                vrf = aim_res.VRF.from_dn(dn)
            except aim_exc.InvalidDNForAciResource:
                raise n_exc.InvalidInput(
                    error_message=('%s is not valid VRF DN' % dn))

            # Check if another address scope already maps to this VRF.
            session = plugin_context.session
            mappings = self._get_address_scope_mappings_for_vrf(session, vrf)
            vrf_owned = False
            for mapping in mappings:
                if mapping.address_scope.ip_version == data['ip_version']:
                    raise n_exc.InvalidInput(
                        error_message=(
                            'VRF %s is already in use by address-scope %s' %
                            (dn, mapping.scope_id)))
                vrf_owned = mapping.vrf_owned

            self._add_address_scope_mapping(
                session, result['id'], vrf, vrf_owned)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.tests import base
from oslo_config import cfg

from gbpservice.contrib.nfp.configurator.agents import firewall as fw
from gbpservice.contrib.nfp.configurator.lib import constants as const
from gbpservice.contrib.nfp.configurator.lib import fw_constants as fw_const
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
    fw_test_data as fo)


class FWaasRpcManagerTestCase(base.BaseTestCase):
    """ Implements test cases for RPC manager methods of firewall agent.

    """

    def __init__(self, *args, **kwargs):
        super(FWaasRpcManagerTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()

    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    @mock.patch(__name__ + '.fo.FakeObjects.conf')
    def _get_FWaasRpcManager_object(self, conf, sc):
        """ Retrieves RPC manager object of firewall agent.

        :param sc: mocked service controller object of process model framework
        :param conf: mocked OSLO configuration file

        Returns: object of firewall's RPC manager and service controller

        """

        agent = fw.FWaasRpcManager(sc, conf)
        return agent, sc

    def _test_event_creation(self, method):
        """ Tests event creation and enqueueing for create/update/delete
        operation of firewall agent's RPC manager.

        Returns: none

        """

        agent, sc = self._get_FWaasRpcManager_object()
        context = {}
        arg_dict = {'context': context,
                    'firewall': self.fo.firewall,
                    'host': self.fo.host}
        with mock.patch.object(sc, 'new_event', return_value='foo') as (
                mock_sc_event), (
                mock.patch.object(sc, 'post_event')) as mock_sc_rpc_event:
            call_method = getattr(agent, method.lower())
            call_method(context, self.fo.firewall, self.fo.host)

            result_dict = arg_dict
            result_dict['firewall'] = {
                'file_path': "/tmp/%s" % (self.fo.firewall['id'])}
            mock_sc_event.assert_called_with(id=method,
                                             data=result_dict, key=None)
            mock_sc_rpc_event.assert_called_with('foo')

    def test_create_firewall_fwaasrpcmanager(self):
        """ Implements test case for create firewall method
        of firewall agent's RPC manager.

        Returns: none

        """

        self._test_event_creation(fw_const.FIREWALL_CREATE_EVENT)

    def test_update_firewall_fwaasrpcmanager(self):
        """ Implements test case for update firewall method
        of firewall agent's RPC manager.

        Returns: none

        """

        self._test_event_creation(fw_const.FIREWALL_UPDATE_EVENT)

    def test_delete_firewall_fwaasrpcmanager(self):
        """ Implements test case for delete firewall method
        of firewall agent's RPC manager.

        Returns: none

        """

        self._test_event_creation(fw_const.FIREWALL_DELETE_EVENT)


class FwaasHandlerTestCase(base.BaseTestCase):
    """ Implements test cases for event handler methods
    of firewall agent.

    """

    def __init__(self, *args, **kwargs):
        super(FwaasHandlerTestCase, self).__init__(*args, **kwargs)
        self.fo = fo.FakeObjects()
        self.ev = fo.FakeEventFirewall()
        self.firewall_rule = {
            'id': 'rule-id', 'action': 'allow',
            'destination_ip_address': '',
            'destination_port': '80',
            'enabled': 'enabled', 'ip_version': 'v4',
            'protocol': 'tcp', 'source_ip_address': '',
            'source_port': '', 'shared': False,
            'position': 1
        }

        self.ev.data['context']['agent_info']['resource'] = 'firewall'

    @mock.patch(__name__ + '.fo.FakeObjects.rpcmgr')
    @mock.patch(__name__ + '.fo.FakeObjects.drivers')
    @mock.patch(__name__ + '.fo.FakeObjects.sc')
    def _get_FwHandler_objects(self, sc, drivers, rpcmgr):
        """ Retrieves event handler object of firewall agent.

        :param sc: mocked service controller object of process model framework
        :param drivers: list of driver objects for firewall agent
        :param rpcmgr: object of configurator's RPC manager

        Returns: object of firewall agents's event handler

        """

        with mock.patch.object(cfg, 'CONF') as mock_cfg:
            mock_cfg.configure_mock(host='foo')
            agent = fw.FWaasEventHandler(sc, drivers, rpcmgr, mock_cfg)
        return agent

    def _test_handle_event(self, rule_list_info=True):
        """ Test handle event method of firewall agent for various
        device configuration operations.

        :param rule_list_info: an atrribute of firewall resource object
        sent from plugin which contains the firewall rules.

        Returns: None

        """

        agent = self._get_FwHandler_objects()
        driver = mock.Mock()

        with mock.patch.object(
            agent.plugin_rpc, 'set_firewall_status') as (
            mock_set_fw_status), (
            mock.patch.object(
                agent.plugin_rpc, 'firewall_deleted')) as (mock_fw_deleted), (
            mock.patch.object(
                driver, fw_const.FIREWALL_CREATE_EVENT.lower())) as (
            mock_create_fw), (
            mock.patch.object(
                driver, fw_const.FIREWALL_UPDATE_EVENT.lower())) as (
            mock_update_fw), (
            mock.patch.object(
                driver, fw_const.FIREWALL_DELETE_EVENT.lower())) as (
            mock_delete_fw), (
            mock.patch.object(
                agent, '_get_driver', return_value=driver)):

            firewall = self.fo._fake_firewall_obj()
            if not rule_list_info:
                firewall_rule_list = []
            else:
                firewall_rule_list = [self.firewall_rule]
            firewall.update({'firewall_rule_list': firewall_rule_list})
            self.ev.data.get('firewall').update(
                {'firewall_rule_list': firewall_rule_list})

            agent_info = self.ev.data['context']['agent_info']
            agent.handle_event(self.ev)
            context = self.fo.neutron_context

            if 'service_info' in self.fo.context:
                self.fo.context.pop('service_info')
            if not rule_list_info:
                if self.ev.id == fw_const.FIREWALL_CREATE_EVENT:
                    mock_set_fw_status.assert_called_with(
                        agent_info,
                        firewall['id'], const.STATUS_ACTIVE, firewall)
                elif self.ev.id == fw_const.FIREWALL_UPDATE_EVENT:
                    mock_set_fw_status.assert_called_with(
                        agent_info,
                        const.STATUS_ACTIVE, firewall)
                elif self.ev.id == fw_const.FIREWALL_DELETE_EVENT:
                    mock_fw_deleted.assert_called_with(
                        agent_info, firewall['id'], firewall)
            else:
                if self.ev.id == fw_const.FIREWALL_CREATE_EVENT:
                    mock_create_fw.assert_called_with(
                        context,
                        firewall, self.fo.host)
                elif self.ev.id == fw_const.FIREWALL_UPDATE_EVENT:
                    mock_update_fw.assert_called_with(
                        context,
                        firewall, self.fo.host)
                elif self.ev.id == fw_const.FIREWALL_DELETE_EVENT:
                    mock_delete_fw.assert_called_with(
                        context,
                        firewall, self.fo.host)

    def test_create_firewall_with_rule_list_info_true(self):
        """ Implements test case for create firewall method
        of firewall agent's event handler with firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_CREATE_EVENT
        self._test_handle_event()

    def test_update_firewall_with_rule_list_info_true(self):
        """ Implements test case for update firewall method
        of firewall agent's event handler with firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_UPDATE_EVENT
        self._test_handle_event()

    def test_delete_firewall_with_rule_list_info_true(self):
        """ Implements test case for delete firewall method
        of firewall agent's event handler with firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_DELETE_EVENT
        self._test_handle_event()

    def test_create_firewall_with_rule_list_info_false(self):
        """ Implements test case for create firewall method
        of firewall agent's event handler without firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_CREATE_EVENT
        self._test_handle_event(False)

    def test_update_firewall_with_rule_list_info_false(self):
        """ Implements test case for update firewall method
        of firewall agent's event handler without firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_UPDATE_EVENT
        self._test_handle_event(False)

    def test_delete_firewall_with_rule_list_info_false(self):
        """ Implements test case for delete firewall method
        of firewall agent's event handler without firewall rules.

        Returns: none

        """

        self.ev.id = fw_const.FIREWALL_DELETE_EVENT
        self._test_handle_event(False)
